import time
import unittest
from domingo.backends.base import QueryError
from domingo.models import AsyncModel
from domingo.fields import Char, Int, List, Set


class TestModel(AsyncModel):

    unique_on = ['char_field']

    char_field = Char()
    int_field = Int()
    list_field = List()
    set_field = Set()
    second_char_field = Char()


class SecondTestModel(TestModel):

    unique_on = ['char_field', 'int_field']


class QueryTests(unittest.TestCase):

    def setUp(self):
        self.instance = TestModel(char_field='hello')
        self.instance.save()

    @classmethod
    def tearDownClass(cls):
        TestModel.store.collection.drop()
        SecondTestModel.store.collection.drop()

    def test_unique_on_works(self):
        "create a second instance and then make sure the count is still one"
        instance = TestModel(char_field='hello')
        instance.save()
        self.assertEqual(
            TestModel.store.collection.count(),
            1
        )

        instance = TestModel(char_field='hello')
        instance._fields['int_field'].set(1)
        instance.save()
        self.assertEqual(
            TestModel.store.collection.count(),
            1
        )

    def test_unique_on_with_get_works(self):
        """
        We require that getting the instance from mongo after creating it, and
        then saving it again does not create duplicates
        """
        instance = TestModel.get(char_field='hello')
        instance.save()
        self.assertEqual(
            TestModel.store.collection.count(),
            1
        )

        instance._fields['int_field'].set(1)
        instance.save()

        # instance = TestModel.get(char_field='hello')
        # instance.save()

        TestModel.create(data=TestModel(char_field='hello'))
        self.assertEqual(
            TestModel.store.collection.count(),
            1
        )

    def test_unique_on_not_instatiated_raises_error(self):
        incomplete_instance = SecondTestModel(char_field='asdfasdf')
        self.assertRaises(QueryError, incomplete_instance.save)
        incomplete_instance = SecondTestModel(int_field=1)
        self.assertRaises(QueryError, incomplete_instance.save)

    def test_default_values_are_not_saved_to_the_collection(self):
        "how about that for a long ass name...."
        document = TestModel.store.collection.find_one({'char_field': 'hello'})
        self.assertTrue(
            'int_field' not in document,
            '"int_field" is in the document! Mongo is saving default values '
            'in the database that are not lists.'
        )
        self.assertTrue(
            'list_field' in document,
            '"list_field" is NOT in the document! Mongo should instatiate '
            'iterables in the database upon creation of the document.'
        )
        self.assertTrue(
            'set_field' in document,
            '"set_field" is NOT in the document! Mongo should instatiate '
            'iterables in the database upon creation of the document.'
        )

    def test_modified_and_created_are_different(self):
        original_doc = TestModel.store.collection.find_one(
            {'char_field': 'hello'}
        )

        instance = TestModel(char_field='hello')
        instance.save()
        doc = TestModel.store.collection.find_one({'char_field': 'hello'})

        self.assertEqual(original_doc['created'], doc['created'])
        self.assertNotEqual(doc['created'], doc['modified'])

        self.assertEqual(instance.modified, doc['modified'])
        self.assertLess(original_doc['modified'], doc['modified'])

    def test_upated_on_single_field_works(self):
        # ensure second_char_field isn't in the docuemnt
        doc = TestModel.store.collection.find_one({'char_field': 'hello'})
        self.assertTrue('second_char_field' not in doc)
        instance = TestModel(
            char_field='hello', second_char_field='yo', int_field=2
        )
        instance.save()
        doc = TestModel.store.collection.find_one({'char_field': 'hello'})
        self.assertEqual(doc['second_char_field'], instance.second_char_field)
        self.assertEqual(doc['int_field'], instance.int_field)

        updated_instance = TestModel(char_field='hello', int_field=10)
        updated_instance.save()
        doc = TestModel.store.collection.find_one({'char_field': 'hello'})
        self.assertEqual(doc['second_char_field'], instance.second_char_field)
        self.assertEqual(doc['int_field'], updated_instance.int_field)

    def test_unique_on_index_is_present(self):
        index_info = TestModel.store.collection.index_information()
        self.assertTrue('unique_on' in index_info)
        self.assertIsNotNone(index_info['unique_on'].get('key'))
        self.assertEquals(index_info['unique_on']['key'], [('char_field', 1)])




from .mongo_backend import MongoStore

import mongomock
from mongomock import Database

M = mongomock.Connection()

class MockStore(MongoStore):
    """
        this is awesome.
        By setting "M" to be the mongomock db, everything 'just' works...
    """


import calendar
import json
import sys


import logging
logger = logging.getLogger(__name__)
time_logger = logging.getLogger('time_logger')

from collections import defaultdict

from datetime import timedelta

from twisted.internet import defer, reactor, task
from twisted.internet.threads import deferToThread
from twisted.web._newclient import ResponseNeverReceived

from .fields import (
    Id, TimeStamp, List, Dict, AbstractBaseField, NonIter, Iter,
    ModelList, ObjectId, AbstractSet, timeseries_namespace
)
from .backends.redis_backend import get_redis
from .connection import get_database
from .query import QueryMixin
from .data_access import DataAccessBase
from .utils import classproperty, utcnow, utcstamp

# leave this in for now as we need to also refactor our YouTube code in the
# same way.
# from apps.platforms.api.youtube import YoutubeApiError


MAX_LIST_LENGTH = 3


class AppCache(object):

    def __init__(self):
        self.model_cache = {}

    def get_model(self, app_label, model_name):
        return self.model_cache.get(
            app_label.lower(), {}
        ).get(model_name.lower())

    def register_model(self, app_label, model_name, model):

        if self.model_cache.get(app_label):
            if self.model_cache.get(app_label.lower()).get(model_name):
                return
            else:
                self.model_cache[app_label.lower()][model_name.lower()] = model
        else:
            self.model_cache[app_label.lower()] = {model_name.lower(): model}


cache = AppCache()
get_model = cache.get_model
register_model = cache.register_model


class MetaModel(object):

    def __init__(self, field_list):
        self.fields = {}
        self.fieldmap = defaultdict(list)
        for name, field in field_list:
            setattr(self, name, field)
            self.fields[name] = field
            self.fieldmap[field.__class__].append(name)
        self.fieldmap = dict(self.fieldmap)


class ModelBaseMeta(type):
    def __new__(meta, classname, bases, classDict):

        app_label = bases[0].__module__.split('.')[-2]

        # first look in the cache and see if we made this already
        existing = get_model(app_label, classname)
        if existing:
            return existing

        # find fields from all base classes of this model
        field_list = []
        for base in bases:
            if hasattr(base, '_meta'):
                field_list += base._meta.fields.items()

        # now remove all fields from this field's class dict
        # so we don't put them directly on the class
        # store them to put them on the _meta
        for key, val in classDict.items():
            if issubclass(val.__class__, AbstractBaseField):
                field_list.append((key, val))
                del classDict[key]

        newmeta = MetaModel(field_list)
        # make the class
        new_class = type.__new__(meta, classname, bases, classDict)
        new_class._meta = newmeta

        # save it
        register_model(app_label, classname, new_class)

        return get_model(app_label, classname)


class AbstractModelBase(object):

    data_access_class = DataAccessBase

    @classproperty
    def namespace(cls):
        return cls.__name__.lower()

    @classproperty
    def store(cls):
        if not hasattr(cls, '_store'):
            cls._store = get_database(cls.storage, cls)
        return cls._store

    @property
    def data(self):
        if hasattr(self, 'data_access_class'):
            if not hasattr(self, '__data_access'):
                self.__data_access = self.data_access_class(self)
            return self.__data_access
        else:
            raise AttributeError('a data_access_class has not been defined for this instance')


class CacheModelBase(AbstractModelBase):

    __metaclass__ = ModelBaseMeta

    storage = 'redis'

    def __init__(self, *args, **kwargs):
        self._fields = {}
        for name, field in self._meta.fields.items():
            FieldClass = field.__class__
            if name in kwargs:
                instance = FieldClass(kwargs[name], owner=self, name=name)
            else:
                instance = FieldClass(owner=self, name=name)
            self._fields[name] = instance
            setattr(self, name, instance)

    def save(self):
        raise NotImplementedError

    def as_dict(self):
        return {
            name: field.value
            for name, field in self._fields.items()
        }


class ModelBase(AbstractModelBase, QueryMixin):
    """
    The objective of this class is to set the main functionaly of the
    interface between the database and the webapp.
    1. Django-like definition of fields on the class
    2.
    """
    __metaclass__ = ModelBaseMeta

    save_list = None
    storage = 'mongo'

    outlet = None
    max_time_series_length = MAX_LIST_LENGTH

    unique_on = []

    @classmethod
    def classfields(cls):
        return cls._meta.fields.items()

    @property
    def unique_key(self):
        return '.'.join([str(getattr(self, name)) for name in self.unique_on])

    @classproperty
    def timeseries_fieldnames(cls):
        "returns a list of strings referencing the field attribute handle"
        if not hasattr(cls, '_timeseries_fieldnames'):
            cls._timeseries_fieldnames = [
                name for name, field in cls.classfields() if field._timeseries
            ]
        return cls._timeseries_fieldnames

    @classproperty
    def redis(cls):
        if not hasattr(cls, '_redis'):
            cls._redis = get_redis(cls.namespace)
        return cls._redis

    def __init__(self, *args, **kwargs):
        self._fields = {}
        # save state on whether or not this model is coming out of the database
        self._read_from_db = False
        if '_read_from_db' in kwargs:
            self._read_from_db = True

        applied_data_lookup = kwargs.pop('_applied_data', None)

        self.context_order = kwargs.pop('_context_order', None)

        setattr(self, 'db_text_score', kwargs.pop('db_text_score', None))
        # time_logger.info('start setup fields')
        self.setup_fields(*args, **kwargs)

        if applied_data_lookup:
            # TODO: this kinda sucks that we need to "ObjectId" the key... it
            # makes it very specific... maybe break this out into a method
            # someday?
            setattr(
                self, applied_data_lookup['namespace'],
                applied_data_lookup['data'].get(ObjectId(self._id.value))
            )

    def __setattr__(self, name, value):
        "ammends setattr so to allow resetting of the local fields"
        # TODO: this may not be needed if we make fields able to somehow
        # correctly set their own values instead of needing "set"
        if not name == '_fields':
            if self._fields.get(name):
                self._fields[name].set(value)

        super(ModelBase, self).__setattr__(name, value)

    def field_init(self, fieldname, field, *args, **kwargs):
        """
        Initializes the field and stores it in self._fields. It also sets the
        fields owner and name, as well as setting the default value to the
        field if no custom value was specified at Model instantiation.
        """
        # time_logger.info('start field %r'%fieldname)
        try:
            FieldClass = field.__class__
            if fieldname in kwargs:
                val = kwargs.get(fieldname)
                newfield = FieldClass(
                    val, owner=self, name=fieldname, **field._original_kwargs
                )
                if kwargs.get('%s_count' % fieldname):
                    newfield.count = kwargs['%s_count' % fieldname]
            else:
                # time_logger.info('last field thing')
                newfield = FieldClass(
                    owner=self, name=fieldname, **field._original_kwargs
                )
            if self._read_from_db:
                newfield.updated = False

            setattr(self, fieldname, newfield.value)
            self._fields[fieldname] = newfield
        except:
            raise
        # time_logger.info('finished field %r'%fieldname)

    def _raise_unique_exception(self):
        if self.__class__.__name__ in ['ModelBase', 'AsyncModel']:
            return
        raise AttributeError(
            '''\n%r must have at least one field specified as unique.
            like this:
                unique_on = ['field_one', 'field_two']

            ''' % self
        ), None, sys.exc_info()[2]

    def setup_fields(self, *args, **kwargs):
        """
        Iterates over all of the attribute names in the model class in order to
        identify which objects are fields and then initialize the model
        instance specific fields (with values). These field instances are kept
        in the dict self._fields.
        """

        for name, field in self._meta.fields.items():
            self.field_init(name, field, *args, **kwargs)

    @property
    def fields(self):
        "Returns a list of the instances fields"
        return self._fields.values()

    @property
    def fieldnames(self):
        return self._fields.keys()

    def _filterFields(self, *args, **kwargs):
        """
        Filters the instances within _fields for the class type specified.
        """
        keep = kwargs.get('keep', True)
        cls = self.__class__
        key = ''.join([c.__name__.lower() for c in args]) + str(keep)
        if not hasattr(cls, key):
            if kwargs.get('exact'):
                ret = {
                    name: field
                    for name, field in self._fields.items()
                    if field.__class__ in args and name not in ['created', '_id']
                }
            else:
                ret = {
                    name: field
                    for name, field in self._fields.items()
                    if issubclass(field.__class__, args) == keep and name not in ['created', '_id']
                }
            setattr(cls, key, ret.keys())
        else:
            fieldnames = getattr(cls, key)
            ret = {name: self._fields[name] for name in fieldnames}
        return ret

    @property
    def list_fields(self):
        return self._filterFields(List, exact=True)

    @property
    def model_list_fields(self):
        return {
            name: field
            for name, field in self._fields.items()
            if isinstance(field, ModelList)
        }

    @property
    def set_fields(self):
        return self._filterFields(AbstractSet)

    @property
    def simple_fields(self):
        return self._filterFields(NonIter, Dict)

    @property
    def timeseries_fields(self):
        now = utcstamp()
        return {
            timeseries_namespace(name): List(
                [now, self._fields[name].value],
                save_on_create=self._fields[name].save_on_create)
            for name in self.timeseries_fieldnames
            if self._fields[name].updated
        }

    @classproperty
    def defaults(cls):
        if not hasattr(cls, '_defaults'):
            defaults = {
                name: field.to_json()
                for name, field in cls.classfields()
                if name not in ['_id'] + cls.unique_on and (isinstance(field, TimeStamp) or issubclass(field.__class__, Iter))
            }
            defaults.update(
                {
                    timeseries_namespace(name): []
                    for name in cls.timeseries_fieldnames
                }
            )
            cls._defaults = defaults
        else:
            defaults = cls._defaults
        return defaults

    @property
    def id(self):
        if not hasattr(self, 'id_cache'):
            self._id_cache = '.'.join(
                [str(self._fields[f]) for f in self.unique_on]
            )
        return self._id_cache

    @property
    def pk(self):
        return self.id

    @property
    def uid(self):
        return '%s.%s' % (self.namespace, self.id)

    def __repr__(self):
        return u'<%s %s>' % (
            self.__class__.__name__,
            self.id,
            # unicode(self.fields)
        )

    def as_dict(self):
        """
        For use in model serialization. It uses the dictionary returned from
        _fields property method (below).
        """
        if not hasattr(self, '__as_dict'):
            data = {
                name: field.to_json() for name, field in self._fields.items()
            }
            self.__as_dict = data
        return self.__as_dict

    def to_json(self):
        "Serializes the model instance"
        return json.dumps(self.as_dict())

    def ordered_values(self, fields=None):
        fields = fields or []
        try:
            return [
                self._fields.get(fieldname).to_json() for fieldname in fields
            ]
        except TypeError, e:
            raise

    def csv(self, fields):
        return '%s\n' % '\t'.join([str(self.fields[f]) for f in fields])

    @classmethod
    def instantiate_list(cls, data, from_json=False, read_from_db=False, offset=0, applied_data=None):
        if from_json:
            return [cls(_read_from_db=read_from_db, _context_order=i+offset, _applied_data=applied_data, **json.loads(r)) for i, r in enumerate(data)]
        return [cls(_read_from_db=read_from_db, _context_order=i+offset, _applied_data=applied_data, **r) for i, r in enumerate(data)]

    @classmethod
    def load_from_cache(cls, key, start, end):
        pipe = cls.redis.pipeline()
        pipe.lrange(key, start, end)
        pipe.llen(key)
        results, count = pipe.execute()

        if count:
            return cls.instantiate_list(results, read_from_db=True, offset=start, from_json=True), count
        return None, None

    @classmethod
    def invalidate_cache(cls, key):
        cls.redis.delete(key)

    @classmethod
    def update_cache(cls, key, match):
        try:
            cls.redis.lset(key, match.context_order, match.to_json())
        except:
            pass


class SimpleModel(ModelBase):
    """
        omits Id, created, and modified fields
    """


class AsyncModel(ModelBase):
    """
    This model must be used for models with one or more "AggregateScore" fields

    It may actually be the best type to use in general as it does not
    alter the base functionality of the standard ModelBase in
    any significant way.
    """

    _id = Id()
    created = TimeStamp(updated=True)
    modified = TimeStamp()

    def __init__(self, *args, **kwargs):
        super(AsyncModel, self).__init__(*args, **kwargs)
        # time_logger.info('begin model')
        self.in_progress = 0
        self.start_time = utcnow()
        self.finished = False
        self.timeout = timedelta(0, kwargs.pop('timeout', 15))
        self.latest_check = utcnow()

        self.processes = []

        self.functions = kwargs.pop('functions', [])
        self.input_data = kwargs.pop('function_args', [None])[0]
        # time_logger.info('end model')

    @property
    def meta_data_key(self):
        """
        This method utilises the microseconds in the time to create a
        unique namespace for the redis cache key-value store.
        """
        epoch_time = calendar.timegm(self.start_time.utctimetuple())
        time_signature = int(
            epoch_time * 1000000 + self.start_time.microsecond
        )
        return '%s.%d' % (self.namespace, time_signature)

    def _set_initial_data(self):
        """
            sets some state info in redis.
            not used for anything currently, but could be used for
            spreading activity across muliple processes
        """
        self.redis.hset(
            self.meta_data_key, 'in_progress', len(self.input_data)
        )

    def run(self, outfile=None):
        """
        sets up some conditions and starts the process for this aggregator
        """

        if hasattr(reactor, 'aggregator_count'):
            reactor.aggregator_count += 1
        else:
            reactor.aggregator_count = 1

        if outfile:
            self.outfile = outfile

        self._set_initial_data()
        self._spawn_chain(tuple(self.functions))

        lc = task.LoopingCall(self.signal_finished, check=True)
        lc.start(self.timeout.total_seconds()/5)

    def _spawn_chain(self, list_of_funcs):
        """
            used at the offset.
            could possibly be replaced by "branch".
        """

        defproc = defer.Deferred()
        for proc in list_of_funcs:
            if isinstance(proc, tuple):
                self._spawn_chain(proc)
            else:
                defproc.addCallback(proc, aggregator=self)
        defproc.callback(self.input_data)

    def branch(self, function=None, args=None,
               kwargs=None, callbacks=None, attempt=0):
        """
            branches callbacks into new deferred processes
        """
        if not callbacks:
            callbacks = []
        if not kwargs:
            kwargs = {}

        kwargs['aggregator'] = self
        if attempt > 0:
            print 'branching: %r attempt %d, args:%r' % (
                function, attempt+1, args)

        D = deferToThread(function, args, **kwargs)

        for c in callbacks:
            D.addCallback(c, **kwargs)

        def errfunction(failure, *args, **kwargs):
            r = failure.trap(
                # YoutubeApiError,
                ResponseNeverReceived,
                defer.CancelledError
            )
            print 'failure:', r
            print 'attempt %d of %r'% (attempt+1, function)
            print 'retrying'
            kwargs.get('aggregator').branch(
                function=function,
                args=args,
                callbacks=callbacks,
                attempt=attempt + 1
            )

        D.addErrback(errfunction, *args, **kwargs)

    def increment_conditions(self, fields=None):
        """
            this is so we can bulk set finishing conditions data across
            multiple fields.
        """
        if not fields:
            fields = {}

        pipe = self.redis.pipeline()
        for field, value in fields.items():
            f = self._fields[field]
            f.increment_condition(value, pipe=pipe)
        pipe.execute()

    def manage_work(self, value, token=None):
        """
            manages how many children are still actively aggregating data
        """
        self.in_progress += value
        if token:
            self.processes.append(token)

    def signal_finished(self, value=1, check=False):
        """
            when an instance spawned by this aggregator
            is done aggregating data, it signals it is finished.
            we then decrement our work counter and see if
            we are all done.
        """
        if check:
            if self.check_time():
                self.finish()
            else:
                print '.',

        self.manage_work(-value)

        if self.in_progress == 0 and self.check_time():
            self.finish()

    def finish(self):
        """
            checks if the reactor has more aggregators running,
            if not, stops the reactor.

            TODO:
            this should probably only be done during testing or script running
            contexts..., how do we know if that is the case?
        """
        if self.finished:
            return

        self.finished = True
        self.write_data()
        if hasattr(reactor, 'aggregator_count'):
            reactor.aggregator_count -= 1

            if reactor.aggregator_count <= 0:
                reactor.stop()

    def write_data(self):
        if not len(self.unique_on):
            self._raise_unique_exception()

        channel_stats = self.get_list(self.processes).values()

        for channel_stat in channel_stats:
            print channel_stat
            aggregation_fields = [
                f for f in channel_stat._fields.values()
                if isinstance(f, fields.AsyncInt)
            ]
            for field in aggregation_fields:
                field.save(force=True)

            print 'finished %s (views c %d, v %d), (count c %d, v %d)' % (
                channel_stat.id,
                channel_stat.channel_views,
                channel_stat.aggregate_views,
                channel_stat.channel_video_count,
                channel_stat.aggregate_video_count)

            out = open(self.outfile, 'a')
            out.write(
                channel_stat.csv([
                    'id',
                    'channel_views',
                    'aggregate_views',
                    'channel_video_count',
                    'aggregate_video_count'
                ])
            )
            out.close()

    def tick(self):
        self.latest_check = utcnow()

    def check_time(self):
        return utcnow() - self.latest_check > self.timeout

    @classmethod
    def post_bulk_create(cls, *args, **kwargs):
        pass




def deriv1(A0, B0, A1, B1, *args):
    """

        # The first derivative equation is (f(t+h) - f(t))/h

        was like this...
        view_diff = timeseries[0][1] - timeseries[1][1]
        time_diff = get_minute(timeseries[0][0] - timeseries[1][0])
        result[0] = view_diff / time_diff

    """

    Adiff = A0 - A1
    Bdiff = B0 - B1
    if Bdiff > 0:
        return Adiff / Bdiff
    return 0


def deriv2(A0, B0, A1, B1, A2, B2, *args):
    """
        # The second derivative equation is
        # (h2f(t+h1) + h1f(t-h2) - (h1+h2)f(t)) / ((h1+h2)h1h2/2)
        h1 = get_minute(timeseries[0][0] - timeseries[1][0])
        h2 = get_minute(timeseries[1][0] - timeseries[2][0])
        fh1 = timeseries[0][1]
        ft = timeseries[1][1]
        fh2 = timeseries[2][1]
        result[1] = h2 * fh1 + h1 * fh2 - (h1 + h2) * ft
        result[1] = result[1] / ((h1 + h2) * h1 * h2 / 2)

    """

    h1 = A0 - A1
    h2 = A1 - A2
    fh1 = B0
    ft = B1
    fh2 = B2
    result = h2 * fh1 + h1 * fh2 - (h1 + h2) * ft
    if any(a == 0 for a in [h1, h2]):
        return 0
    return result / ((h1 + h2) * h1 * h2 / 2)


funcs = [
    None,
    deriv1,
    deriv2
]


def calculate_derivatives(timeseries, attribute_pairs, depth=2):
    """
        attribute_pairs:
            [(view_count, time), (sub_count, time)]etc...

    """

    for A, B in attribute_pairs:
        args = []
        for t in timeseries[:max(2, len(timeseries))]:
            args.extend([getattr(t, A).to_json(), getattr(t, B).to_json()])
        results = []
        for order in range(1, depth+1):
            if len(args) >= 2 + 2*order:
                timeseries[0].set_derivatives(A, B, funcs[order](*args), order)
    return timeseries



from django.db import IntegrityError
from django.db.models import Manager, signals
from django.db.models.fields import Field
from django.db.models.query import QuerySet, ValuesListQuerySet, ValuesQuerySet
from django.utils import six
from domingo.exceptions import QueryError

from collections import defaultdict


class DomingoForeignKey(object):
    """
        class MyModel(django.db.models.Model):
            ...
            ref_field = django.db.models.CharField(max_length=...)
            doc = DomingoForeignKey(
                MyDomingoModel, unique_on_field='ref_field'
            )
    """
    description = "Domingo foreign key"
    include_fields = None
    exclude_fields = None

    # Field flags
    auto_created = False
    concrete = False
    editable = False
    hidden = False

    is_relation = True
    many_to_many = False
    many_to_one = True
    one_to_many = False
    one_to_one = False
    related_model = None

    def __unicode__(self):
        return u'.'.join(['hello', ])

    def __init__(self, cls, *args, **kwargs):
        self.domingo_model = cls
        self.unique_on = cls.unique_on

        if 'include_fields' in kwargs:
            self.include_fields = kwargs.pop('include_fields')
        if 'exclude_fields' in kwargs:
            self.exclude_fields = kwargs.pop('exclude_fields')

        self.editable = False
        self.rel = None
        self.column = None

        # there should be a regular django field for each field listed in the
        # unique_on attribute on the referenced domingo model
        try:
            self.ref_fieldnames = {
                name: kwargs.pop(name) for name in self.unique_on
            }
        except KeyError:
            raise AttributeError(
                'Specify a reference field for each field listed in %s.'
                'unique_on' % (self.__class__.__name__, cls.__name__)
            )
        super(DomingoForeignKey, self).__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super(DomingoForeignKey, self).deconstruct()
        # Only include kwarg if it's not the default
        if self.separator != ",":
            kwargs['separator'] = self.separator
        return name, path, args, kwargs

    def contribute_to_class(self, cls, name):
        self.name = name
        self.model = cls
        # django model
        self.django_model = cls
        cls._meta.add_field(self, virtual=True)
        if hasattr(cls._meta, 'domingo_foreign_key'):
            raise AttributeError('Only one DFK per model')
        # Only run pre-initialization field assignment on non-abstract models
        if not cls._meta.abstract:
            signals.pre_init.connect(self.instance_pre_init, sender=cls)
        cls._meta.domingo_foreign_key = name

        # Connect myself as the descriptor for this field
        setattr(cls, name, self)

    def instance_pre_init(self, signal, sender, args, kwargs, **_kwargs):
        if self.name in kwargs:
            domingo_instance = kwargs.pop(self.name)
            for fieldname, ref_fieldname in self.ref_fieldnames.items():
                kwargs[ref_fieldname] = getattr(domingo_instance, fieldname)
            setattr(self, self.cache_key, domingo_instance)

    @property
    def cache_key(self):
        return '_%s_domingo_instance' % self.name

    def __get__(self, instance, instance_type=None):
        if instance is None:
            raise AttributeError('')
        if not hasattr(instance, self.cache_key):
            search_opts = {
                fieldname: getattr(instance, ref_fieldname)
                for fieldname, ref_fieldname in self.ref_fieldnames.items()
            }
            if self.include_fields:
                search_opts['include'] = self.include_fields
            if self.exclude_fields:
                search_opts['exclude'] = self.exclude_fields
            try:
                domingo_instance = self.domingo_model.get(**search_opts)
            except QueryError, e:
                raise IntegrityError(str(e))
            setattr(instance, self.cache_key, domingo_instance)
        return getattr(instance, self.cache_key)

    def __set__(self, instance, value):
        if not isinstance(value, self.domingo_model):
            raise TypeError()
        for fieldname, ref_fieldname in self.ref_fieldnames.items():
            setattr(instance, ref_fieldname, getattr(value, fieldname))
        setattr(instance, self.cache_key, value)


class DomingoQuerySet(QuerySet):
    """
    Overrides the __iter__ method on the queryset so to allow the addition of
    the outlet instance into the model (relative to the outlet_key field)
    """

    _domingo_filter_opts = None
    _domingo_order_keys = None
    _domingo_instances_list = None

    def select_related(self, *args, **kwargs):
        if hasattr(self.model._meta, 'domingo_foreign_key'):
            if self.model._meta.domingo_foreign_key in args:
                self.requires_domingo = True
        return super(DomingoQuerySet, self).select_related(
            *[
                arg
                for arg in args
                if arg != getattr(self.model._meta, 'domingo_foreign_key', None)
            ],
            **kwargs
        )

    @property
    def requires_domingo(self):
        if not hasattr(self, '_requires_domingo'):
            query = self.query
            if isinstance(query.select_related, bool):
                self._requires_domingo = query.select_related
            else:
                self._requires_domingo = False
        return self._requires_domingo

    @requires_domingo.setter
    def requires_domingo(self, value):
        self._requires_domingo = value

    def django_unique_key(self, instance):
        unique_on = self.dfk_field.domingo_model.unique_on
        ref_fieldnames = self.dfk_field.ref_fieldnames
        return '.'.join([
            unicode(getattr(instance, ref_fieldnames[name]))
            for name in unique_on
        ])

    def __iter__(self):
        self._fetch_all()

        is_data = isinstance(self, (ValuesQuerySet, ValuesListQuerySet))
        if is_data or not self.requires_domingo:
            for res in self._result_cache:
                yield res
        else:
            self._fetch_domingo_data()

            if self._domingo_order_keys:
                # logic for returning instances in the order specified by the
                # domingo outlets

                instance_lookup = defaultdict(list)
                for django_instance in self._result_cache:
                    unique_key = self.django_unique_key(django_instance)
                    instance_lookup[unique_key].append(
                        django_instance
                    )

                all_tokens = set(instance_lookup.keys())
                domingo_tokens = []

                for domingo_instance in self._domingo_instances_list:
                    token = domingo_instance.unique_key
                    instances = instance_lookup.get(token, [])
                    for instance in instances:
                        setattr(
                            instance,
                            '_%s_domingo_instance' % self.dfk_field.name,
                            domingo_instance
                        )
                        yield instance
                    domingo_tokens.append(token)

            else:

                all_tokens = []
                domingo_tokens = []

                for django_instance in self._result_cache:
                    unique_key = self.django_unique_key(django_instance)
                    domingo_instance = self._domingo_instances_lookup.get(
                        unique_key
                    )
                    all_tokens.append(unique_key)
                    if domingo_instance:
                        setattr(
                            django_instance,
                            '_%s_domingo_instance' % self.dfk_field.name,
                            domingo_instance
                        )
                        domingo_tokens.append(domingo_instance.unique_key)
                        yield django_instance

    def _get_domingo_opts(self):
        name = self.dfk_field.name
        ret = defaultdict(list)
        ref_fieldnames = self.dfk_field.ref_fieldnames
        for instance in self._result_cache:
            for fieldname, ref_fieldname in ref_fieldnames.items():
                ret[fieldname + "__in"].append(
                    getattr(instance, ref_fieldname)
                )
        setattr(self, '_%s_opts' % name, dict(ret))
        opts = getattr(self, '_%s_opts' % name)
        self._domingo_filter_opts = getattr(self, '_domingo_filter_opts') or {}
        opts.update(self._domingo_filter_opts)
        return opts

    @property
    def dfk_field(self):
        if not hasattr(self, '_dfk_field'):
            self._dfk_field = self.model._meta.get_field(
                self.model._meta.domingo_foreign_key
            )
        return self._dfk_field

    def _fetch_domingo_data(self):
        """
        Get the outlets via a mongo cursor and assign them to the 'hidden'
        attribute _domingo_instances_lookup. _domingo_instances_lookup is a
        dict that returns the outlet instance for a given outlet_token.
        """
        opts = self._get_domingo_opts()
        DomingoModel = self.dfk_field.domingo_model
        count = DomingoModel.filter(**opts).count()
        _lim = 40
        self._domingo_instances_lookup = {}
        self._domingo_instances_list = []

        key_parts = None
        if self._domingo_order_keys:
            key = self._domingo_order_keys[0]
            key_parts = (key, int(not key.startswith('-')))

        if count > 2 * _lim:
            pages = count / _lim
            for idx in range(0, pages):
                cursor = DomingoModel.filter(**opts)
                if key_parts:
                    cursor = cursor.sort(*key_parts)
                cursor = cursor.limit(_lim).skip(idx * _lim)
                if self._domingo_order_keys:
                    for outlet in cursor:
                        self._domingo_instances_list.append(outlet)
                else:
                    self._domingo_instances_lookup.update(
                        {
                            outlet.unique_key: outlet
                            for outlet in cursor
                        }
                    )
        else:
            cursor = DomingoModel.filter(**opts)
            if key_parts:
                cursor = DomingoModel.filter(**opts)
                self._domingo_instances_list = cursor.sort(*key_parts)
            else:
                self._domingo_instances_lookup.update(
                    {
                        outlet.unique_key: outlet for outlet in cursor
                    }
                )

    def __getitem__(self, k):
        """
        Retrieves an item or slice from the set of results.
        """
        if not isinstance(k, (slice,) + six.integer_types):
            raise TypeError
        assert ((not isinstance(k, slice) and (k >= 0)) or
                (isinstance(k, slice) and (k.start is None or k.start >= 0) and
                 (k.stop is None or k.stop >= 0))), \
            "Negative indexing is not supported."

        # if self._result_cache is not None:
            # return self._result_cache[k]

        if isinstance(k, slice):
            qs = self._clone()
            if k.start is not None:
                start = int(k.start)
            else:
                start = None
            if k.stop is not None:
                stop = int(k.stop)
            else:
                stop = None
            qs.query.set_limits(start, stop)
            return list(qs)[::k.step] if k.step else qs

        qs = self._clone()
        qs.query.set_limits(k, k + 1)
        return list(qs)[0]

    def filter(self, *args, **kwargs):
        _domingo_filter_opts = {}
        dfk_fieldname = self.dfk_field.name
        domingo_model = self.dfk_field.domingo_model
        for key, value in kwargs.items():
            key_parts = key.split('__')
            if dfk_fieldname in key_parts:
                idx = key_parts.index(dfk_fieldname)
                lk = '__'.join(key_parts[idx + 1:])
                if isinstance(value, domingo_model) and not lk:
                    for name in domingo_model.unique_on:
                        _domingo_filter_opts[name] = getattr(value, name)
                elif lk == 'in' and hasattr(value, '__iter__'):
                    for name in domingo_model.unique_on:
                        _domingo_filter_opts[name + '__in'] = []
                    for val in value:
                        for name in domingo_model.unique_on:
                            try:
                                _domingo_filter_opts[name + '__in'].append(
                                    getattr(val, name)
                                )
                            except AttributeError:
                                pass
                else:
                    _domingo_filter_opts[lk] = value
                del kwargs[key]
        ###################################################################
        # Ensures that the filter will always generate sql
        # TODO: use the ids from this query to later retrieve the outlet
        # data
        if _domingo_filter_opts:
            dfk_field = self.dfk_field
            DomingoModel = self.dfk_field.domingo_model
            domingo_instances = DomingoModel.filter(
                include=DomingoModel.unique_on,
                **_domingo_filter_opts
            )
            ref_fieldnames = dfk_field.ref_fieldnames
            for fieldname, ref_fieldname in ref_fieldnames.items():
                kwargs.update(
                    {
                        ref_fieldname + '__in': [
                            getattr(instance, fieldname)
                            for instance in domingo_instances
                        ]
                    }
                )
        #######################################################################
        qs = super(DomingoQuerySet, self).filter(*args, **kwargs)
        qs._domingo_filter_opts = dict(_domingo_filter_opts)
        return qs

    def values_list(self, *fields, **kwargs):
        if self.requires_domingo:
            for field in fields:
                if self.dfk_field.name in field.split('__'):
                    raise NotImplementedError(
                        'values_list has not been impletemented for '
                        'domingo models'
                    )
        return super(DomingoQuerySet, self).values_list(*fields, **kwargs)

    def values(self, *fields, **kwargs):
        if self.requires_domingo:
            for field in fields:
                if self.dfk_field.name in field.split('__'):
                    raise NotImplementedError(
                        'values has not been impletemented for domingo '
                        'models'
                    )
        return super(DomingoQuerySet, self).values(*fields, **kwargs)

    def order_by(self, *args, **kwargs):
        pattern = '%s__' % self.dfk_field.name
        keys = [
            key for key in args
            if key.startswith(pattern) or key.startswith('-' + pattern)
        ]
        _domingo_order_keys = [
            key.replace(pattern, '') for key in keys
        ]
        if len(_domingo_order_keys) > 1:
            raise ValueError(
                "Domingo models can only be sorted by one field at a time. "
                "'%s' keys found." % _domingo_order_keys
            )

        args = tuple(set(args) - set(keys))
        self._domingo_order_keys = _domingo_order_keys
        self.requires_domingo = bool(_domingo_order_keys)
        return super(DomingoQuerySet, self).order_by(*args, **kwargs)

    def _clone(self):
        clone = super(DomingoQuerySet, self)._clone()
        clone._domingo_order_keys = self._domingo_order_keys
        clone._domingo_filter_opts = self._domingo_filter_opts
        clone.requires_domingo = self.requires_domingo
        # clone._domingo_instances_list = self._domingo_instances_list
        return clone

    def exists(self):
        django_exists = super(DomingoQuerySet, self).exists()
        if self.requires_domingo and self._domingo_filter_opts:
            if not self._result_cache:
                self._fetch_all()
            DomingoModel = self.dfk_field.domingo_model
            opts = self._get_domingo_opts()
            domingo_exists = bool(
                DomingoModel.filter(**opts).count()
            )
            return django_exists and domingo_exists
        return django_exists

    def count(self):
        if self.requires_domingo and self._domingo_filter_opts:
            return len(list(self))
        return super(DomingoQuerySet, self).count()


class DomingoManager(Manager):

    def get_queryset(self):
        return DomingoQuerySet(self.model)



from .support import calculate_derivatives
from logging import getLogger


logger = getLogger('time_logger')


class QueryMixin(object):

    def pre_save(self):
        """
            called before save
        """
        pass

    @classmethod
    def filter(cls, *args, **kwargs):
        return cls.store.filter(*args, **kwargs)

    @classmethod
    def all(cls):
        return cls.store.all()

    @classmethod
    def search(cls, keyword, **kwargs):

        return cls.store.filter(
            text_search=keyword,
            projection = {'db_text_score': {'$meta': "textScore"}},
            **kwargs).sort("textScore")

    @classmethod
    def get(cls, **kwargs):
        """
            needs a kwarg like:
                outlet_token=qlrejfnweiruybf
        """
        return cls.store.get(**kwargs)

    def get_derivative(self, attribute_pair, depth=2):
        if hasattr(self, 'time_series') and len(self.time_series) >= depth:
            return calculate_derivatives(
                self.time_series, [attribute_pair], depth
            )
        return 0

    @classmethod
    def get_latest_maxima_list(cls, exclude=None):
        obj = cls()
        pipe = obj.redis.pipeline()

        outfields = []

        for name, field in obj.number_fields.items():
            if exclude is not None and name in exclude:
                continue
            outfields.append(name)
            pipe.zrevrangebyscore(
                '%s.%s' % (obj.namespace, field.key),
                'INF', 0, 0, 1, withscores=True
            )
        data = pipe.execute()
        if data:
            return {f: data[i][0][1] for i, f in enumerate(outfields)}

    @classmethod
    def create(cls, data=None, pipe=None):
        data = data if isinstance(data, list) else [data]
        cls.store.save(data=data, created=True)
        return data

    @classmethod
    def update(cls, queryset, **kwargs):
        for m in queryset:
            for key, val in kwargs.items():
                getattr(m,key).set(val)
            m._id.updated = True
        cls.store.save(data=queryset, created=True, fields=kwargs.keys())

        return queryset

    def save(self, fields=None):
        self.store.save(data=[self], fields=fields)
        return self



import json
import os
import redis
import sys

from .base import DatabaseBackendBase, QuerySet, QueryError
from collections import defaultdict

try:
    from django.conf import settings
    from django.core.exceptions import ImproperlyConfigured
    has_django = True
except ImportError, e:
    has_django = False


if has_django:
    try:
        REDIS_HOST = getattr(settings, 'DOMINGO_REDIS_HOST', 'localhost')
        REDIS_PORT = getattr(settings, 'DOMINGO_REDIS_PORT', '6379')
        REDIS_DB = getattr(settings, 'DOMINGO_REDIS_DB', 10)
        REDIS_PW = getattr(settings, 'DOMINGO_REDIS_PW', None)
    except ImproperlyConfigured, e:
        has_django = False


if not has_django:
    REDIS_HOST = os.environ.get('DOMINGO_REDIS_HOST', 'localhost')
    REDIS_PORT = os.environ.get('DOMINGO_REDIS_PORT', '6379')
    REDIS_DB = os.environ.get('DOMINGO_REDIS_DB', 10)
    REDIS_PW = os.environ.get('DOMINGO_REDIS_PW', None)


R = redis.StrictRedis(
    host=REDIS_HOST, port=REDIS_PORT, db=REDIS_DB, password=REDIS_PW
)


class RedisQuerySet(QuerySet):
    def count(self):
        pass


def get_redis(key=None):
    return R


class RedisStore(DatabaseBackendBase):

    queryset_class = RedisQuerySet

    def __init__(self, model, **kwargs):
        self.connection = get_redis(model.namespace)
        self.model = model
        self.namespace = model.namespace

    def commit_bulk_save(self):
        pipe = R.pipeline()
        for k, v in self.bulk_query:
            pipe.lpush(k, v.to_json())
        return pipe.execute()

    def commit_bulk_get(self, time_series=0):
        pipe = R.pipeline()
        for k in self.bulk_query:
            pipe.lrange(
                k, 0, time_series
            )

        data = pipe.execute()

        return [json.loads(d[0]) for d in data]

    def get(self, key):
        d = self.connection.lindex(key, 0)
        return d

    def unique_key(self, instance):
        names = []
        valid = True
        for name in instance.unique_on:
            field = getattr(instance, name)
            names.append(str(name))
            valid &= field.updated
        if not valid:
            raise QueryError(
                "%r does not have a complete unique_on key" % instance
            ), None, sys.exc_info()[2]
        return ".".join(names)

    @staticmethod
    def model_keys(keys):
        return [key for key in keys if len(key.split(':')) <= 2]

    @staticmethod
    def field_keys(keys):
        return [key for key in keys if len(key.split(':')) > 2]

    def all(self):
        """
        an all query method based on using : for delimiting the model, instance,
        field values in the key. The majority of the model is to be saved in a
        hash and the list fields extend the key to include the field name.
        The idea here is that the has can be updated with the list structures
        after grabbing the data.
        """
        keys = self.connection.keys(self.namespace + ":*")
        ret = defaultdict(dict)
        pipe = self.connection.pipeline()
        model_keys = self.model_keys(keys)
        field_keys = self.field_keys(keys)
        for key in model_keys:
            pipe.hgetall(key)
        model_data = pipe.execute()
        for idx, key in enumerate(model_keys):
            namespace, unique_key = key.split(":")
            ret[unique_key].update(
                dict(zip(self.model.unique_on, unique_key.split(".")))
            )
            ret[unique_key].update(model_data[idx])
        pipe = self.connection.pipeline()
        for key in field_keys:
            name, method = key.split('.')
            getattr(pipe, method)(key)
        pipe.execute()


    def save(self, data=None, sequential=True, created=False, fields=None):
        """
        data: list of model instances
        sequential: True if order is important, False otherwise
        created: True if the model instances are to have been created for
            the first time
        """
        bulkop = self.connection.pipe()

        if not fields:
            fields = []

        invalid_instances = []
        for instance in data:
            unique_key = {}
            valid = True

            for name in instance.unique_on:
                field = getattr(instance, name)
                unique_key[name] = field.value
                valid &= field.updated
                # valid is false if any of the unique_on fields were not
                # instantiated with an actual value

            if not valid:
                invalid_instances.append(instance)
                continue

            unique_key = '%s.%s' % (
                self.model.namespace, '.'.join(unique_key.values()))

            print unique_key

            bulkop.set(unique_key, instance.defaults)

            for name, field in instance.simple_fields.items():
                if field.updated or name == 'modified':
                    bulkop.hset(unique_key, name, field.value)

            if instance.timeseries_fields:
                for key, value in instance.timeseries_fields.items():
                    if not created or value.save_on_create:
                        bulkop.lpush('%s.%s' % (unique_key, key), value)

            if instance.set_fields:
                for key, value in instance.set_fields.items():
                    bulkop.sadd('%s.%s' % (unique_key, key), value)

            if instance.list_fields:
                for key, field in instance.list_fields.items():
                    if field.updated:
                        bulkop.lpush('%s.%s' % (unique_key, key), field.value)

            return
            """
            if instance.model_list_fields:
                # get the full models...
                for key in instance.unique_on:


                self.model.filter()


                model_list_data = {
                    key: {'$each': [m.as_dict() for m in field.value]}
                    for key, field in instance.model_list_fields.items() if field.updated
                }
                if model_list_data:
                    op = {'$addToSet': model_list_data}

                    if '$addToSet' in operations:
                        operations['$addToSet'].update(op)
                    else:
                        operations.update(op)
            # print json.dumps(operations, indent=5)
            bulkop.find(unique_key).upsert().update(operations)

        if invalid_instances:
            raise QueryError(
                'Could not save the data, %d out of the %d instances were '
                'invalid.' % (
                    len(invalid_instances), len(data)
                )
            )

        try:
            result = bulkop.execute()
        except BulkWriteError, e:
            logger.error(str(e), exc_info=True, extra={'data': e.details})
            raise
        return result

        # self.model.update_min_max_fields(data)
        """



import unittest

from domingo import models
from domingo.fields import List, Char, Dict, Int
from domingo.utils import utcnow, utcstamp


class YoutubeStats(models.AsyncModel):
    """
        mostly just used for testing.
    """
    unique_on = ['outlet_token', 'outlet_type']

    outlet_token = Char()
    outlet_type = Char()
    view_count = List()
    video_count = List()
    subscriber_count = List()
    comment_count = List()
    median_view_count = List()
    days_since_last_video = List()

    metadata = Dict()


class ModelBaseTests(unittest.TestCase):

    def setUp(self):
        self.model_instance = models.SimpleModel()
        self.async_instance = models.AsyncModel()
        self.yts_instance = YoutubeStats()

    # @unittest.skip('Skipping for now. Runs strangely sometimes.')
    def test_instantiation_time(self):
        fake_data = dict(
            outlet_token='asdlfkjasldfkj',
            outlet_type=int(),
            view_count=range(2),
            video_count=range(2),
            subscriber_count=range(2),
            comment_count=range(2),
            median_view_count=range(2),
            days_since_last_video=range(2)
        )

        diffs = []
        for i in range(500):
            start = utcstamp()
            YoutubeStats(**fake_data)
            finish = utcstamp()
            diff = finish - start
            diffs.append(diff)
        avg_diff = sum(diffs) / 500.

        limits = [0.01, 0.0025, 0.002, 0.0015, 0.001, 0.00053]
        for limit in limits:
            self.assertLess(
                avg_diff, limit,
                "Instantiation of a model takes too long. Limit was %f secs."
                "Took %f seconds" % (limit, diff)
            )

    def test_namespace(self):
        "Tests that inheritance doesn't effect the namespace"
        self.assertEqual(models.SimpleModel.namespace, 'simplemodel')
        self.assertEqual(self.model_instance.namespace, 'simplemodel')

        self.assertEqual(models.AsyncModel.namespace, 'asyncmodel')
        self.assertEqual(self.async_instance.namespace, 'asyncmodel')

        self.assertEqual(YoutubeStats.namespace, 'youtubestats')
        self.assertEqual(self.yts_instance.namespace, 'youtubestats')

    def test_model_instance_specific_fields(self):
        """
        test that the two (class field and instance field) are not the same
        object
        """
        # it also tests that ModelBase has the Timestamp field added by default
        self.assertFalse(
            self.async_instance.created is models.AsyncModel._meta.created
        )

    def test_model_instance_fields_getattr(self):
        """
        The value of all fields within self._fields should be
        equivalent to the attribute
        named on the instance
        """
        created = self.async_instance._fields['created']
        self.assertTrue(
            created.value is self.async_instance.created
        )

    def test_setattr(self):
        """
        You should be able to set the value of a local field by 'normal' means,
        i.e. model.field = value
        """
        time = utcnow()
        self.async_instance.created = time
        self.assertNotEqual(
            models.AsyncModel._meta.created, self.async_instance.created
        )
        self.assertEqual(self.async_instance.created, time)


from django.db.models import get_model
from domingo.fields import NonIter, List, AbstractSet


class DjangoFKey(NonIter):

    instance = None
    encoder = staticmethod(int)

    def to_json(self):
        return int(self.value)

    def load(self):
        if not self.instance and self.value and self.django_model:
            Model = get_model(*self.django_model.split('.'))
            self.instance = Model.objects.get(pk=self.value)
        return self.instance


class DjangoFKList(List, AbstractSet):
    index = True

    def __init__(self, *args, **kwargs):
        self.safe_model = kwargs.get('model').replace('.', '__')
        super(DjangoFKList, self).__init__(*args, **kwargs)

    def encoder(self, *args):

        value = super(DjangoFKList, self).encoder(*args)
        for i, v in enumerate(value):
            if not type(v) == dict:
                value[i] = {self.safe_model: v}
        return value

    def prep_query_value(self, value):
        return {self.safe_model: value}

    def get_queryset(self):
        Model = get_model(*self.django_model.split('.'))
        self.queryset = Model.objects.get(
            pk__in=[d.values()[0] for d in self.value]
        )
        return self.queryset



import calendar
import random
from datetime import datetime


def hash96(seed=None):
    """
    Returns a 24 hexidecimal hash of the seed given. If no seed is given then
    it returns a random hash based on the internal clock.
    ipython:
        %timeit hash96('hello')
        10000 loops, best of 3: 21.8 microseconds per loop on a macbook air
    N.B. This function has been checked across OS platforms and cpus... yeah
    it's python...
    """

    def _hash(bits=96):
        assert bits % 8 == 0
        required_length = bits / 8 * 2
        s = hex(random.getrandbits(bits)).lstrip('0x').rstrip('L')
        if len(s) < required_length:
            return _hash(bits)
        else:
            return s
    random.seed(seed)

    return _hash()


class ClassPropertyDescriptor(object):

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls=None):
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj, value):
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func):
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


def classproperty(func):
    if not isinstance(func, (classmethod, staticmethod)):
        func = classmethod(func)

    return ClassPropertyDescriptor(func)


utcnow = datetime.utcnow


def utcstamp(
        timestamp=None,
        utcnow=datetime.utcnow,
        utcfromtimestamp=datetime.utcfromtimestamp,
        timegm=calendar.timegm
    ):
    "returns seconds since epoch in utc time [float]"
    if isinstance(timestamp, datetime):
        now = timestamp
    else:
        if not timestamp or timestamp is None:
            now = utcnow()
        else:
            now = utcfromtimestamp(float(timestamp))
    epoch_time = timegm(now.utctimetuple())
    return epoch_time + now.microsecond / 1000000.



from .backends.mongo_backend import MongoStore
from .backends.redis_backend import RedisStore
from .backends.mock_backend import MockStore
# from .backends.rethink_backend import RethinkStore


def get_database(storage, model):

    if storage == 'mongo':
        return MongoStore(model)
    if storage == 'redis':
        return RedisStore(model)
    if storage == 'mock':
        return MockStore(model)
    # if storage == 'rethink':
        # return RethinkStore(namespace)





class QueryError(Exception):
    "Exception to handle when dealing with queries"
    pass


class DatabaseBackendBase(object):

    def start_bulk(self):
        self.bulk_query = []
        return self.bulk_query

    def commit_bulk_create(self):
        raise NotImplementedError

    def clean_token(self, token):
        return token.split('.')[1]

    def add_filter(self, key, value):
        raise NotImplemented('need to implement add filter for this class')


class QuerySet(object):
    """
        allows database query results to be
        lazily instantiated
    """

    default_query = None

    def __init__(self, query=None, model=None, db=None, offset=0, apply_data=None):
        self._result_cache = None
        self.query = query or self.default_query
        self.model = model
        self.db = db
        self.applied_data=apply_data

        self.offset = offset

    @property
    def results(self):
        """
            only hit the database if we don't have
            a result cache.  otherwise return that instead
        """
        if not self._result_cache:
            try:
                self._result_cache = self.model.instantiate_list(
                    self.query,
                    read_from_db=True,
                    offset=self.offset,
                    from_json=False,
                    applied_data=self.applied_data
                )
            except Exception, e:
                raise
                print 'r:',r
        return self._result_cache

    def __iter__(self):
        return iter(self.results)

    def __getitem__(self, index):
        if self.results:
            return self.results[index]
        return None

    def __len__(self):
        if self.results:
            return len(self.results)
        return 0

    def count(self):
        return len(self.results)

    def as_list(self):
        return [o.as_dict() for o in self.results]



import unittest
from domingo.utils import classproperty


class TestModel(object):
    "A class for use only in this test file"

    _test = 'yo'

    @classproperty
    def test(cls):
        return 'hello'

    @classproperty
    def test2(cls):
        return cls._test


class ClassProperyTests(unittest.TestCase):

    def test_class_usage(self):

        self.assertEqual('hello', TestModel.test)
        self.assertEqual('yo', TestModel.test2)

    def test_instance_usage(self):

        self.assertEqual('hello', TestModel().test)
        self.assertEqual('yo', TestModel().test2)

    def test_inheritance_usage(self):

        class Inherit(TestModel):
            _test = 'works!'

        self.assertEqual('hello', Inherit().test)
        self.assertEqual('works!', Inherit().test2)



import logging
import sys
from bson.objectid import ObjectId
from .backends.redis_backend import get_redis
from .utils import utcstamp


logger = logging.getLogger(__name__)
time_logger = logging.getLogger('time_logger')


def timeseries_namespace(name):
    return name + '_timeseries'


class AbstractBaseField(object):
    has_dependencies = False
    updated = False

    reserved = dir(object) + ['set', 'append', 'to_json', 'reserved']

    def __init__(self, *args, **kwargs):
        # time_logger.info('start field init')
        self.name = None
        self.unique = False
        self.save_on_create = True
        self._original_kwargs = kwargs
        self._timeseries = kwargs.get('time_series', False)

        for k, v in kwargs.items():
            if k not in self.reserved:
                setattr(self, k, v)

        self.set(*args)
        # time_logger.info('end field init')

    @property
    def timeseries(self):
        if self._timeseries:
            return self.owner[timeseries_namespace(self.name)]
        return None

    def __repr__(self):
        return unicode(self.value)

    def __eq__(self, other):
        if self.value is None:
            if other is None:
                return True
            else:
                return False
        elif other is None:
            return False
        else:
            value = self.value
            if issubclass(other.__class__, AbstractBaseField):
                other_value = other.value
                return not value < other_value and not other_value < value
            else:
                return not value < other and not other < value

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        if isinstance(other, (int, long, float)):
            return self.value > other
        if issubclass(other.__class__, AbstractBaseField):
            return self.value > other.value

    def __ge__(self, other):
        return self > other or self == other

    def __le__(self, other):
        return self < other or self == other

    def __sub__(self, other):
        if issubclass(other.__class__, AbstractBaseField):
            return self.value - other.value
        else:
            return self.value - other

    def __add__(self, other):
        if issubclass(other.__class__, AbstractBaseField):
            return self.value + other.value
        else:
            return self.value + other

    def __radd__(self, other):
        if issubclass(other.__class__, AbstractBaseField):
            return other.value + self.value
        else:
            return other + self.value

    def __mul__(self, other):
        if issubclass(other.__class__, AbstractBaseField):
            return self.value * other.value
        else:
            return self.value * other

    def __div__(self, other):
        if issubclass(other.__class__, AbstractBaseField):
            return self.value / other.value
        else:
            return self.value / other

    def __int__(self):
        return int(self.value)

    def __hash__(self):
        return self.value

    def __neg__(self):
        return self.value * -1

    @property
    def key(self):
        """
        Returns a unique name to be used in a cache store.
        We can change this method to return a shorter value if we want
        """
        return self.name

    def to_json(self):
        return self.value

    def get_default(self):
        if hasattr(self, 'default'):
            return self.default
        else:
            return self.encoder()

    def set(self, *args, **kwargs):
        try:
            self.value = self.encoder(args[0])
            self.updated = True
        except IndexError:
            self.value = self.get_default()

    def prep_query_value(self, value):
        return value

    def is_not_empty(self):
        return self.value != self.encoder() or (
            hasattr(self, 'default') and self.value == self.default)


###############################################################################
# Delineate field by whether they are iterable or not.
class Iter(AbstractBaseField):

    iterable = True

    def __len__(self):
        return len(self.value)

    def __iter__(self):
        return iter(self.value)


class NonIter(AbstractBaseField):

    iterable = False


class Bool(NonIter):

    encoder = staticmethod(bool)

    def to_json(self):
        return int(self.value)


class Id(NonIter):

    encoder = staticmethod(ObjectId)

    def to_json(self):
        return str(self.value)


class FKey(NonIter):

    encoder = staticmethod(ObjectId)

    def to_json(self):
        return str(self.value)


class Char(NonIter):

    def encoder(self, *args):
        if args:
            return unicode(args[0]).encode("utf-8")


class TimeStamp(NonIter):

    encoder = staticmethod(utcstamp)

    def __eq__(self, other):

        return utcstamp(self.value) == utcstamp(other)


class Number(NonIter):

    _min = None
    _max = None

    def __repr__(self):
        return str(self.value)


class Float(Number):

    encoder = staticmethod(float)


class Int(Number):

    encoder = staticmethod(int)


class AsyncNumber(Number):
    """
    The value of this score is the result of aggregating data that may
    come from multiple sources or api calls.  This field contains
    mechanisms for dealing the collection and incrementing of such score
    data.
    """

    # datetimes
    started = None
    finished = None

    def __init__(self, *args, **kwargs):
        """
        The following enforces the setting of the instance attributes that
        are needed to specify the finishing condition.
        """
        condition_reqs = ('condition_type', 'condition_key')
        reqs = {req: req not in kwargs for req in condition_reqs}
        if any(reqs.values()):  # checks if any of the reqs are not included
            raise AttributeError(
                "%s.%s needs to have a %s specified" % (
                    self.owner, self.name,
                    ' & '.join([req for req, met in reqs.items() if met])
                )
            ), None, sys.exc_info()[2]
        super(AsyncNumber, self).__init__(*args, **kwargs)

    @property
    def aggregate_key(self):
        """
        The key where we will store the aggregate value in progress.

        N.B. self.owner is defined on the instance specific initialisation of
        the field. (so is self.name)... self.owner refers to the model instance
        that initialised the field (i.e. the model in which the field lives)
        Goto AbstractModelBase._setup_fields for more info.
        """
        return '%s.%s.%s' % (self.namespace, self.owner.id, self.key)

    def get_data(self):
        """
        Gets the temporary value stored in redis and re-sets the value of this
        field.
        Returns a dict containing the value
        """
        connection = get_redis(self.namespace)
        self._data = connection.hgetall(self.aggregate_key)
        self.set(self._data.get('value', 0))
        return self._data

    @property
    def is_complete(self):
        """
        Checks if the aggregation process is complete by checking the localized
        data store against the logic associated to the type of condition is
        associated to the field.
        """
        data = self.get_data()
        actual_count = int(data.get(self.condition_key))
        if self.condition_type == 'decrement':
            return actual_count <= 0
        else:
            return actual_count >= 0

    def add_data(self, value, samples=1):
        """
        Call to asynchronously add data to the temporary redis store and to
        update the condition_key by the given number of samples.
        """
        connection = get_redis(self.namespace)
        pipe = connection.pipeline()
        pipe.hincrby(self.aggregate_key, 'value', value)
        if self.condition_type == 'decrement':
            pipe.hincrby(self.aggregate_key, self.condition_key, -samples)
        else:
            pipe.hincrby(self.aggregate_key, self.condition_key, samples)
        data = pipe.execute()
        self.set(data[0])

    def increment_condition(self, value, pipe=None):
        if not pipe:
            pipe = get_redis(self.namespace)
        pipe.hincrby(self.aggregate_key, self.condition_key, value)

    def save(self, force=False):
        """
        Saves the final value to a redis key-value store specific to the
        field's model. It saves the value and the model id.

        At the moment this function is only used in the write_data method in
        the ScoreModel class.
        """
        self.get_data()
        connection = get_redis(self.namespace)
        # if force is true then is_complete will not be evaluated
        if not force and not self.is_complete:
            return False
        try:
            # stores the model.id in a sorted set ordered by the value.
            connection.zadd(
                '%s.%s' % (self.owner.namespace, self.key),
                self.value,
                self.owner.id
            )
        except Exception, e:
            print 'could not save %r for %s.  value was %r - %s' % (
                self, self.owner.id, self.value, str(e)
            )

        self.owner.save()
        return self.value


class AsyncInt(AsyncNumber, Int):

    namespace = 'sc-agg'



class AbstractSet(object):
    "A hack to prevent the import of django dependent fields"
    pass


def to_list(*args):
    if not args:
        return []
    value = args[0]
    return list(value) if isinstance(value, (list, tuple, set)) else [value,]


class FKList(Iter, AbstractSet):

    encoder = staticmethod(to_list)

    def to_json(self):
        return [str(element) for element in self.value]



class List(Iter):

    encoder = staticmethod(to_list)

    def append(self, value):
        self.value.append(value)

    def __getitem__(self, index):
        return self.value[index]


class ModelList(List):

    def encoder(self, *args):
        if args:
            if type(args[0]) == list:
                if type(args[0][0]) == dict:
                    return [self.model_class(**v) for v in args[0]]
                elif type(args[0][0]) == self.model_class:
                    return args[0]
            else:
                if type(args[0]) == dict:
                    return self.model_class(**args[0])
                elif type(args[0]) == self.model_class:
                    return args[0]
        return []

    def set(self, *args, **kwargs):
        try:
            self.value = self.encoder(args[0])
            if self.order_by:
                self.value = sorted(
                    self.value, key=lambda m: getattr(m, self.order_by),
                    reverse=True
                )
            self.updated = True
        except IndexError:
            self.value = self.get_default()

    def to_json(self):
        return [m.to_json() for m in self.value]


class Set(Iter, AbstractSet):

    encoder = staticmethod(to_list)

    def append(self, value):
        self.value = list(set(self.value).add(value))

    def get(self, name):
        if name in self.value:
            return name
        else:
            raise ValueError(
                '%r.%s does not contain the value you are looking for...' % (
                    self.owner, self.name
                )
            ), None, sys.exc_info()[2]


class Dict(Iter):

    encoder = staticmethod(dict)

    def items(self):
        return self.value.items()

    def get(self, key):
        return self.value.get(key)



from domingo import __version__
from setuptools import setup, find_packages


setup(
    author="hangarunderground",
    author_email="domingo@reelio.com",
    name="domingo",
    packages=find_packages(exclude=['domingo/tests/*']),
    version=__version__,
    url="https://github.com/hangarunderground/domingo",
    download_url=(
        "https://github.com/hangarunderground/domingo/tarball/" + __version__
    ),
    description="A django-like modelling interface for various schemaless dbs",
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
    ],
    keywords=["django", "schema-less", "models", "data"],
    install_requires=['django', 'pymongo', 'redis', 'mongomock', 'twisted'],
    extras_require={'rethink': ['rethinkdb', ], }
)



import logging
import os
import rethinkdb as rethink
from .base import DatabaseBackendBase

logger = logging.getLogger(__name__)


try:
    from django.conf import settings
    from django.core.exceptions import ImproperlyConfigured
    has_django = True
except ImportError, e:
    has_django = False


if has_django:
    try:
        RETHINK_HOST = getattr(settings, 'DOMINGO_RETHINK_HOST', 'localhost')
        RETHINK_PORT = getattr(settings, 'DOMINGO_RETHINK_PORT', '28015')
        RETHINK_DB = getattr(settings, 'DOMINGO_RETHINK_DB', 'rethink_domingo_test')
    except ImproperlyConfigured, e:
        has_django = False


if not has_django:
    RETHINK_HOST = os.environ.get('DOMINGO_RETHINK_HOST', 'localhost')
    RETHINK_PORT = os.environ.get('DOMINGO_RETHINK_PORT', '28015')
    RETHINK_DB = os.environ.get('DOMINGO_RETHINK_DB', 'rethink_domingo_test')

rtconnection = rethink.connect(
    host=RETHINK_HOST, port=RETHINK_PORT
)
try:
    rethink.db_create(RETHINK_DB).run(rtconnection)
except rethink.RqlRuntimeError:
    pass
    # rtconnection.close()


def connect(f):
    def wrapper(self, *args, **kwargs):
        self._setup_connection()
        return f(self, *args, **kwargs)
    return wrapper


class RethinkStore(DatabaseBackendBase):

    tables = {}

    def __init__(self, namespace, **kwargs):
        self.namespace = namespace
        self._setup_connection()
        try:
            rethink.db(settings.RETHINK_DB).table_create(
                self.namespace, primary_key='outlet_token'
            ).run(self.connection)
        except rethink.RqlRuntimeError:
            logger.debug('table %s already exists', self.namespace)
        try:
            rethink.db(settings.RETHINK_DB).table_create(
                self.namespace+'_scores'
            ).run(self.connection)
            rethink.db(settings.RETHINK_DB).table(
                self.namespace+'_scores'
            ).index_create('outlet_token').run(self.connection)
        except rethink.RqlRuntimeError:
            logger.debug('table %s already exists', self.namespace)

    def _setup_connection(self):
        self.connection = rethink.connect(
            host=settings.RETHINK_HOST, port=settings.RETHINK_PORT
        )

    @connect
    def commit_bulk_save(self):

        rethink.db(settings.RETHINK_DB).table(self.namespace).insert(
            [data.as_dict() for id, data in self.bulk_query],
            upsert=True
            ).run(self.connection)

        rethink.db(settings.RETHINK_DB).table(self.namespace+'_scores').insert(
            [data.as_dict() for id, data in self.bulk_query],
            # durability='soft',
            ).run(self.connection)

        self.connection.close()

    @connect
    def commit_bulk_get(self, time_series=0):
        if not len(self.bulk_query):
            return []

        query = rethink.db(settings.RETHINK_DB).table(
            self.namespace).get_all(
                *[k.split('.')[-1] for k in self.bulk_query]
            ).eq_join(
                'outlet_token', rethink.db(
                    settings.RETHINK_DB
                ).table(self.namespace+'_scores')
            ).distinct().zip().order_by(rethink.desc('time'))

        data = query.run(self.connection)
        self.connection.close()
        return data

    @connect
    def get(self, key):
        key = key.split('.')[-1]
        data = rethink.db(settings.RETHINK_DB).table(
            self.namespace+'_scores').filter(
                'outlet_token', key).eq_join(
                    'outlet_token',
                    rethink.db(settings.RETHINK_DB).table(
                        self.namespace)
                    ).order_by(
                        rethink.desc('time')
                    ).limit(1).zip().run(self.connection)
        if data:
            return data[0]



from unittest import TestCase
from domingo.fields import AbstractSet
from domingo.contrib.django.fields import DjangoFKList


class DjangoContribTests(TestCase):

    def test_DjangoFKList_is_a_set_field(self):
        self.assertTrue(issubclass(DjangoFKList, AbstractSet))



import logging
import os
import sys

from .base import DatabaseBackendBase, QuerySet, QueryError
from bson.objectid import ObjectId
from pymongo import MongoClient, DESCENDING, ASCENDING, TEXT, MongoReplicaSetClient
from pymongo.errors import BulkWriteError, OperationFailure


logger = logging.getLogger(__name__)


ALREADY_INDEXED = []

try:
    from django.conf import settings
    from django.core.exceptions import ImproperlyConfigured
    has_django = True
except ImportError, e:
    has_django = False


DEFAULT_DB_NAME = 'domingo'
if has_django:
    try:
        MONGO_REPLICA_SET_URI = getattr(settings, 'MONGO_REPLICA_SET_URI', None)
        MONGO_REPLICA_SET_NAME = getattr(settings, 'MONGO_REPLICA_SET_NAME', None)
        MONGO_DB_NAME = getattr(settings, 'MONGO_DB_NAME', DEFAULT_DB_NAME)
        TEST_MODE = getattr(settings, 'TEST_MODE', True)
    except ImproperlyConfigured, e:
        has_django = False


if not has_django:
    MONGO_REPLICA_SET_URI = os.environ.get('DOMINGO_MONGO_REPLICA_SET_URI')
    MONGO_REPLICA_SET_NAME = os.environ.get('DOMINGO_MONGO_REPLICA_SET_NAME')
    MONGO_DB_NAME = os.environ.get('DOMINGO_MONGO_DB_NAME', DEFAULT_DB_NAME)
    TEST_MODE = os.environ.get('DOMINGO_TEST_MODE', False)

if MONGO_REPLICA_SET_URI and MONGO_REPLICA_SET_NAME:
    M = MongoReplicaSetClient(
        MONGO_REPLICA_SET_URI, replicaSet=MONGO_REPLICA_SET_NAME
    )
else:
    M = MongoClient()


class MongoQuerySet(QuerySet):

    def count(self):
        return self.query.count()

    def sort(self, key, order=ASCENDING):
        if isinstance(key, list):
            self.query = self.query.sort(key)
        else:
            if key.startswith('-'):
                key = key.strip('-')
                order = DESCENDING
            self.query = self.query.sort(key, order)
        return self

    def skip(self, skip):
        self.offset = skip
        self.query = self.query.skip(skip)
        return self

    def limit(self, limit):
        self.query = self.query.limit(limit)
        return self

    def inflate(self, field, **kwargs):
        """
            used when a model has a field that is a reference
            to another model stored in the database.

            this will parse the current model's data
            and return a second queryset which contains
            the result of the query on the 2nd model
            referred to by the 1st model's field
        """
        meta_field = getattr(self.model._meta, field)
        if hasattr(meta_field, 'model_class'):
            field_model = meta_field.model_class

            if hasattr(field_model, 'refers_to'):
                inflate_model = field_model.refers_to['model']
                inflate_field = field_model.refers_to['on_field']

                parent_model_refs = {}
                lists = [getattr(o, field) for o in self]

                # lists = [a list of references to a submodel for each result
                # in this queryset]

                for olist in lists:
                    # for each list of model references

                    # add to the global list of IDs we will pull from the database...
                    for o in olist:
                        parent_model_refs[ObjectId(str(o.fields[inflate_field]))] = {
                            field : o[field].value
                        for field in o.applied_fields
                        }
                return inflate_model.filter(
                    _id__in=parent_model_refs.keys(),
                    offset_query=self.offset,
                    apply_data={
                        'data': parent_model_refs,
                        'namespace': self.model.namespace
                    },
                    **kwargs
                )

        return self


class MongoStore(DatabaseBackendBase):

    queryset_class = MongoQuerySet

    def __init__(self, model, **kwargs):
        self.db = M[MONGO_DB_NAME]
        self.offset_query = 0
        self.applied_data = kwargs.pop('applied_data', {})
        self.model = model
        self.namespace = model.namespace
        self.collection = self.db[self.namespace]
        if 'unique_on' not in self.collection.index_information():
            self.collection.create_index(
                [(name, 1) for name in self.model.unique_on],
                unique=True, name='unique_on', background=True
            )
        if self.namespace not in ALREADY_INDEXED:
            for name, field in self.model.classfields():
                self.set_index(name, field)

            if not TEST_MODE:
                if hasattr(self.model, 'text_index_fields'):
                    try:
                        self.collection.ensure_index(
                            [(f, TEXT) for f in self.model.text_index_fields],
                            name="text_index",
                            weights=getattr(
                                self.model, 'text_index_weights', {}
                            )
                            )
                    except OperationFailure as e:
                        # print e
                        # there is already a different index in place...
                        try:
                            # it is probably the original one where we did all
                            # fields...
                            self.collection.drop_index("$**_text")
                        except OperationFailure as e:
                            logger.error('could not drop index: %r', e)
                        try:
                            # or maybe we changed the options on the index?
                            self.collection.drop_index("text_index")
                        except OperationFailure as e:
                            logger.error('could not drop index: %r', e)

                        self.collection.ensure_index(
                            [(f, TEXT) for f in self.model.text_index_fields],
                            name="text_index",
                            weights=getattr(
                                self.model, 'text_index_weights', {}
                            )
                        )
            ALREADY_INDEXED.append(self.namespace)
        else:
            'indexing done'

    def set_index(self, name, field):
        if hasattr(field, 'unique') and field.unique:
            self.collection.ensure_index(name, unique=True, sparse=True)
        elif hasattr(field, 'index') and field.index:
            self.collection.ensure_index(name)
        # if hasattr(field, 'text_index') and field.text_index:
            # self.text_indexed[name] = 'text'

    def get_operator(self, operator):
        try:
            # TODO: test these
            return {
                'in': '$in',
                'gte': '$gte',
                'gt': '$gt',
                'lt': '$lt',
                'lte': '$lte',
                'not': '$not',
                'ne': '$ne',
                'eq': '$eq'
            }[operator]
        except KeyError:
            raise KeyError(
                'cannot resolve %s for query in %s' % (
                    operator, self.__class__.__name__
                )
            )

    def add_filter(self, key, value):
        out = {}
        if '__' in key:
            field, operator = key.split('__')
            field_class = getattr(self.model._meta,field)
            value = field_class.prep_query_value(value)

            if '__' in operator:
                modifier, primary = operator.split('__')
                out[field] = {
                    self.get_operator(modifier): {
                        self.get_operator(primary): value
                    }
                }
                return out
            elif operator == 'not':
                out[field] = {'$not': {'$eq': value}}
            else:
                out[field] = {self.get_operator(operator): value}
            return out
        field_class = getattr(self.model._meta, key)
        value = field_class.prep_query_value(value)

        out[key] = value
        return out

    def parse_conditions(self, **kwargs):
        """
        Returns a dictionary with the field name as the key and a dictionary
        of operators and values as the field name's value
        """
        conditions = {}
        for key, val in kwargs.items():
            if key == 'text_search':
                conditions['$text'] = {'$search': val}
            else:
                field = ''
                if '__' in key:
                    field, operator = key.split('__')
                if field in conditions:
                    conditions[field].update(self.add_filter(key, val)[field])
                else:
                    conditions.update(self.add_filter(key, val))
        return conditions

    def parse_projections(self, include, exclude, slice_args, elem_match):
        ret = {}
        for field in include:
            ret[field] = 1
        for field in exclude:
            ret[field] = 0

        if slice_args:
            self.offset_query = slice_args.get('start', 0)
            start = slice_args.get('start', 0)
            end = slice_args.get('end')
            if end:
                slice = [start, end]
            else:
                slice = start
            ret[slice_args.get('field')] = {'$slice': slice}
        if elem_match:
            for group in elem_match:
                for field in group['fields']:
                    # print field, group
                    ret[field] = {
                        '$elemMatch': {group['object']: group['value']}
                    }
        # print json.dumps(ret, indent=5)
        return ret

    def all(self):
        return self.queryset_class(
            self.collection.find(), self.model, self.db,
            offset=self.offset_query, apply_data=self.applied_data
        )

    def filter(self, raw=None, include=None, exclude=None, **kwargs):
        include = include or []
        exclude = exclude or []

        self.offset_query = kwargs.pop('offset_query', 0)
        self.applied_data = kwargs.pop('apply_data', self.applied_data)

        slice_args = kwargs.pop('slice', None)
        elem_match = kwargs.pop('elem_match', None)

        projection = kwargs.pop('projection', None)
        if not projection:
            projection = self.parse_projections(
                include, exclude, slice_args, elem_match
            )
        conditions = self.parse_conditions(**kwargs)

        if raw:
            conditions.update(raw)

        # print 'query:'
        # try:
        #     print json.dumps(conditions, indent=5)
        #     print json.dumps(projection, indent=5)
        # except:
        #     pass

        if projection:
            documents = self.collection.find(conditions, projection)
        else:
            documents = self.collection.find(conditions)

        return self.queryset_class(
            documents, self.model, self.db,
            offset=self.offset_query, apply_data=self.applied_data
        )

    def rawFilter(self, query, projection=None):
        if projection:
            documents = self.collection.find(
                query, projection,
                offset=self.offset_query, apply_data=self.applied_data
            )
        else:
            documents = self.collection.find(query)

        return self.queryset_class(
            documents, self.model, self.db,
            offset=self.offset_query, apply_data=self.applied_data
        )

    def get(self, include=None, exclude=None, **kwargs):

        conditions = self.parse_conditions(**kwargs)
        document = self.collection.find(conditions)
        count = document.count()

        if count > 1:
            raise QueryError(
                '%s.get returned too many documents: %s' % (self.model, kwargs)
            ), None, sys.exc_info()[2]
        if count == 1:
            return self.queryset_class(document, self.model, self.db).results[0]
        else:
            raise QueryError(
                '%s.object not found: %s' % (self.model, kwargs)
            ), None, sys.exc_info()[2]

    def save(self, data=None, sequential=True, created=False, fields=None):
        """
        data: list of model instances
        sequential: True if order is important, False otherwise
        created: True if the model instances are to have been created for
            the first time
        """

        if sequential:
            bulkop = self.collection.initialize_ordered_bulk_op()
        else:
            bulkop = self.collection.initialize_unordered_bulk_op()

        invalid_instances = []

        if not fields:
            fields = self.model.defaults.keys()

        for index, instance in enumerate(data):
            instance.pre_save()
            # print [name for name,f in instance.classfields() if f.updated]

            unique_key = {}
            not_empty = True
            for name in instance.unique_on:
                field = instance._fields[name]
                unique_key[name] = field.value
                # print name, field.value, field.is_not_empty()
                not_empty &= field.is_not_empty()
                # field is not_empty if any of the unique_on fields were not
                # instantiated with an actual value
            if not not_empty and instance._fields['_id'].value and instance._fields['_id'].updated:
                unique_key = {'_id': instance._id}


            if not not_empty:
                invalid_instances.append(instance)
                continue

            bulkop.find(unique_key).upsert().update(
                {'$setOnInsert': instance.defaults}
            )

            operations = {}

            operations.update({
                '$set': {
                    name: field.value
                    for name, field in instance.simple_fields.items()
                    if field.updated or name == 'modified' and name in fields
                }
            })

            if instance.timeseries_fields:
                operations.update({
                    '$push': {
                        key: value.value # value is a List() class, so we want its value...
                        for key, value in instance.timeseries_fields.items()
                        if not created or value.save_on_create and key in fields
                    }
                })

            if instance.set_fields:
                set_data = {
                    key: {'$each': field.value}
                    for key, field in instance.set_fields.items()
                    if field.updated and key in fields
                }
                if set_data:
                    operations.update({'$addToSet': set_data})

            if instance.list_fields:
                list_data = {
                    key: {'$each': field.value}
                    for key, field in instance.list_fields.items()
                    if field.updated and key in fields
                }
                if list_data:
                    op = {'$push': list_data}
                    if '$push' in operations:
                        operations['$push'].update(op)
                    else:
                        operations.update(op)

            if instance.model_list_fields:
                model_list_data = {
                    key: {'$each': [m.as_dict() for m in field.value]}
                    for key, field in instance.model_list_fields.items()
                    if field.updated and key in fields
                }
                if model_list_data:
                    op = {'$addToSet': model_list_data}

                    if '$addToSet' in operations:
                        operations['$addToSet'].update(op)
                    else:
                        operations.update(op)

                    for key, val in model_list_data.items():
                        operations['$set']['%s_count' % key] = len(val['$each'])

            for key, val in operations.items():
                if not val:
                    del operations[key]

            bulkop.find(unique_key).upsert().update(operations)

        if invalid_instances:
            raise QueryError(
                'Could not save the data, %d out of the %d instances were '
                'invalid.' % (
                    len(invalid_instances), len(data)
                )
            )

        try:
            result = bulkop.execute()
        except BulkWriteError, e:
            logger.error(str(e), exc_info=True, extra={'data': e.details})
            raise
        return result

        # self.model.update_min_max_fields(data)
    def update(self, filter=None, update=None):
        if not filter:
            raise QueryError('Please filter models to update')
        self.collection.update(filter, update)



import unittest
from domingo.fields import AsyncInt, Int, Float, List, Char


class FieldTests(unittest.TestCase):
    """
    Tests the various interactions amongst fields and between python data
    types and fields
    """

    def setUp(self):
        self.int_field = Int(1)

        self.float_field = Float(3.5)

        self.list_data = range(4)
        self.list_field = List(self.list_data)

        self.char_data = 'hello, this is a test'
        self.char_field = Char(self.char_data)

    def test_eq(self):
        self.assertTrue(self.list_field, self.list_data)
        self.assertEqual(self.list_field, self.list_data)

        self.assertFalse(self.int_field == self.float_field)
        self.assertTrue(self.int_field == 1.)

        self.assertEqual(self.char_field, self.char_data)
        self.assertEqual(self.char_field, Char('hello, this is a test'))
        self.assertTrue(self.char_field == self.char_data)

    def test_ne(self):
        self.assertFalse(self.list_field != self.list_data, 'list fail')
        self.assertNotEqual(self.list_field, range(10), 'list fail')

        self.assertFalse(self.int_field != 1., 'int fail')
        self.assertTrue(self.int_field != self.float_field, 'int fail')

        self.assertNotEqual(self.char_field, 'wrong')
        self.assertNotEqual(self.char_field, Char('breaking shit are we??'))
        self.assertFalse(self.char_field != self.char_data)
        self.assertTrue(self.char_field != 'fail')

    def test_multiply(self):
        self.assertEqual(self.int_field * 2, 2)
        self.assertEqual(self.int_field * self.float_field, 3.5)

    def test_divide(self):
        self.assertEqual(self.int_field / 2, 0)
        self.assertEqual(self.int_field / 2., 0.5)
        self.assertEqual(self.int_field / self.float_field, 1 / 3.5)

    def test_add(self):
        self.assertEqual(self.int_field + 1., 2)
        self.assertEqual(self.int_field + self.float_field, 4.5)

    def test_radd(self):
        self.assertEqual('before' + self.char_field, 'before' + self.char_field)
        self.assertNotEqual(self.char_field, Char('breaking shit are we??'))

    def test_subtract(self):
        self.assertEqual(self.int_field - 1., 0)
        self.assertEqual(self.int_field - self.float_field, -2.5)

    def test_lt(self):
        self.assertFalse(self.int_field < 2)
        self.assertFalse(self.int_field < 0)
        self.assertTrue(self.int_field < self.float_field)

    def test_gt(self):
        self.assertFalse(self.int_field > 2)
        self.assertTrue(self.int_field > 0)
        self.assertFalse(self.int_field > self.float_field)

    def test_ge(self):
        self.assertTrue(self.int_field >= 0)
        self.assertTrue(self.int_field >= 1.)
        self.assertFalse(self.int_field >= self.float_field)

    def test_le(self):
        self.assertFalse(self.int_field <= 0)
        self.assertTrue(self.int_field <= 1.)
        self.assertTrue(self.int_field <= self.float_field)


class AsyncIntTests(unittest.TestCase):

    def setUp(self):
        self.base_kwargs = {'owner': 'ho', 'name': 'stuff'}

    def test_init_w_nothing(self):
        self.assertRaises(
            AttributeError, AsyncInt, **self.base_kwargs
        )

    def test_init_w_condition_type(self):
        self.assertRaises(
            AttributeError, AsyncInt, condition_type="decrement",
            **self.base_kwargs
        )

    def test_init_w_condition_key(self):
        self.assertRaises(
            AttributeError, AsyncInt, condition_key='hello',
            **self.base_kwargs
        )

    def test_init_w_all_reqs(self):
        AsyncInt(
            condition_type="decrement", condition_key="hi", **self.base_kwargs
        )
